CLIS_DESC = [("environment", ".environment_cli.cli")]

FLOW_DECORATORS_DESC = [
    ("conda_base", ".conda.conda_flow_decorator.CondaFlowDecorator"),
    ("pip_base", ".conda.conda_flow_decorator.PipFlowDecorator"),
]
STEP_DECORATORS_DESC = [
    ("conda", ".conda.conda_step_decorator.CondaStepDecorator"),
    ("pip", ".conda.conda_step_decorator.PipStepDecorator"),
]
ENVIRONMENTS_DESC = [("conda", ".conda.conda_environment.CondaEnvironment")]

__mf_promote_submodules__ = ["conda"]



import json
import os
import re
import subprocess
from typing import Dict, List, Optional

from metaflow import Step
from metaflow_extensions.netflix_ext.plugins.conda.conda import Conda
from metaflow_extensions.netflix_ext.plugins.conda.env_descr import (
    EnvID,
    ResolvedEnvironment,
    TStr,
)
from metaflow_extensions.netflix_ext.plugins.conda.utils import arch_id

_deps_parse = re.compile(r"([^<>=!~]+)(.*)")
_ext_parse = re.compile(r"([-_\w]+)\(([^)]+)\)")

name_to_pkg = {"netflix-ext": "metaflow-netflixext"}


def parse_deps(deps: str) -> List[TStr]:
    deps = deps.split(";")
    result = []
    for dep in deps:
        match = _deps_parse.match(dep)
        if match is None:
            raise ValueError("Dependency %s is not a valid constraint")
        pkg, version = match.groups()
        pkg_with_category = pkg.split("::")
        if len(pkg_with_category) > 1:
            category = pkg_with_category[0]
            pkg = pkg_with_category[1]
        else:
            category = "conda"
        version = version.lstrip("=")
        result.append(TStr(category, "%s==%s" % (pkg, version)))
    return result


def parse_channels(channels: Optional[str]) -> List[TStr]:
    if not channels:
        return []

    result = []
    channels = channels.split(",")
    for c in channels:
        channel_with_category = c.split("::")
        if len(channel_with_category) > 1:
            category = channel_with_category[0]
            c = channel_with_category[1]
        else:
            category = "conda"
        result.append(TStr(category, c))
    return result


def req_id_from_spec(python: str, deps: str, channels: str) -> str:
    deps = parse_deps(deps)
    channels = parse_channels(channels)
    deps.append(TStr("conda", "python==%s" % python))
    return ResolvedEnvironment.get_req_id(deps, channels)


def env_id_from_step(step: Step) -> EnvID:
    conda_env_id = step.task.metadata_dict.get("conda_env_id")
    if conda_env_id:
        conda_env_id = json.loads(conda_env_id)
        if isinstance(conda_env_id, type([])) and len(conda_env_id) == 3:
            return EnvID(
                req_id=conda_env_id[0], full_id=conda_env_id[1], arch=arch_id()
            )
        else:
            raise ValueError(
                "%s ran with a version of Metaflow that is too old for this functionality"
                % str(step)
            )
    else:
        raise ValueError("%s did not run with a Conda environment" % str(step))


def local_instances_for_req_id(conda: Conda, req_id: str):
    created_envs = conda.created_environments()  # This returns all MF environments
    local_instances = {}  # type: Dict[str, List[str]]
    for env_id, present_list in created_envs.items():
        if env_id.req_id != req_id:
            continue
        local_instances[env_id.full_id] = present_list
    return local_instances


def download_mf_version(executable: str, version_str: str):
    def _install_pkg(pkg: str, ver: str):
        try:
            subprocess.check_call(
                [
                    executable,
                    "-m",
                    "pip",
                    "install",
                    "-t",
                    ".",
                    "--no-deps",
                    "%s==%s" % (pkg, ver),
                ]
            )
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                "Could not install version '%s' of '%s': %s" % (ver, pkg, e.stderr)
            )

    s = version_str.split("+", 1)
    _install_pkg("metaflow", s[0])
    if len(s) == 1:
        return
    # We now install the other packages, they are in the format name(ver);name(ver)...
    s = s[1].split(";")
    for pkg_desc in s:
        m = _ext_parse.match(pkg_desc)
        if not m:
            raise ValueError("Metaflow extension '%s' is not a valid format" % pkg_desc)
        pkg_name, pkg_version = m.groups()
        pkg = name_to_pkg.get(pkg_name)
        if pkg is None:
            raise ValueError("Metaflow extension '%s' is not known" % pkg_name)
        _install_pkg(pkg, pkg_version)



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

import importlib
import json
import os
import platform
import re
import shutil
import sys
import tempfile

from itertools import chain

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

from metaflow.plugins.datastores.local_storage import LocalStorage
from metaflow.datastore.flow_datastore import FlowDataStore
from metaflow.datastore.task_datastore import TaskDataStore
from metaflow.debug import debug
from metaflow.decorators import StepDecorator
from metaflow.extension_support import EXT_PKG
from metaflow.flowspec import FlowSpec
from metaflow.graph import FlowGraph
from metaflow.metadata import MetaDatum
from metaflow.metadata.metadata import MetadataProvider
from metaflow.metaflow_config import (
    CONDA_REMOTE_COMMANDS,
    get_pinned_conda_libs,
)
from metaflow.metaflow_environment import (
    InvalidEnvironmentException,
    MetaflowEnvironment,
)
from .env_descr import (
    TStr,
    EnvID,
    EnvType,
    ResolvedEnvironment,
)
from metaflow.plugins.env_escape import generate_trampolines
from metaflow.unbounded_foreach import UBF_CONTROL, UBF_TASK
from metaflow.util import get_metaflow_root

from metaflow_extensions.netflix_ext.vendor.packaging.utils import canonicalize_version

from .utils import arch_id, conda_deps_to_pip_deps, merge_dep_dicts
from .conda import Conda


class CondaStepDecorator(StepDecorator):
    """
    Specifies the Conda environment for the step.

    Information in this decorator will augment any
    attributes set in the `@conda_base` flow-level decorator. Hence
    you can use `@conda_base` to set common libraries required by all
    steps and use `@conda` to specify step-specific additions or replacements.
    Information specified in this decorator will augment the information in the base
    decorator and, in case of a conflict (for example the same library specified in
    both the base decorator and the step decorator), the step decorator's information
    will prevail.

    Parameters
    ----------
    name : Optional[str]
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this step. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : Optional[str]
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the pathspec, you can use `@{}` values and
        environment variables will be used to substitute.
    libraries : Optional[Dict[str, str]]
        Libraries to use for this step. The key is the name of the package
        and the value is the version to use (default: `{}`). Note that versions can
        be specified either as a specific version or as a comma separated string
        of constraints like "<2.0,>=1.5".
    channels : Optional[List[str]]
        Additional channels to search
    pip_packages : Optional[Dict[str, str]]
        Same as libraries but for pip packages
    pip_sources : Optional[List[str]]
        Same as channels but for pip sources
    python : Optional[str]
        Version of Python to use, e.g. '3.7.4'. If not specified, the current version
        will be used.
    fetch_at_exec : bool, default False
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this step to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    disabled : bool, default False
        If set to True, disables Conda.
    """

    name = "conda"
    TYPE = "conda"

    defaults = {
        "name": None,
        "pathspec": None,
        "libraries": {},
        "channels": [],
        "pip_packages": {},
        "pip_sources": [],
        "python": None,
        "fetch_at_exec": None,
        "disabled": None,
    }  # type: Dict[str, Any]

    conda = None  # type: Optional[Conda]
    _local_root = None  # type: Optional[str]

    def is_enabled(self, ubf_context: Optional[str] = None) -> bool:
        if ubf_context == UBF_CONTROL:
            return False
        return not next(
            x
            for x in [
                self._self_disabled(),
                self._base_attributes["disabled"],
                False,
            ]
            if x is not None
        )

    def is_fetch_at_exec(self, ubf_context: Optional[str] = None) -> bool:
        if ubf_context == UBF_CONTROL:
            return False
        return next(
            x
            for x in [
                self.attributes["fetch_at_exec"],
                self._base_attributes["fetch_at_exec"],
                False,
            ]
            if x is not None
        )

    @property
    def env_ids(self) -> List[EnvID]:
        # Note this returns a list because initially we had support to specify
        # architectures in the decorator -- keeping for now as this is still valid code
        debug.conda_exec(
            "Requested for step %s: deps: %s; sources: %s"
            % (self._step_name, str(self.step_deps), str(self.source_deps))
        )
        return [
            EnvID(
                req_id=ResolvedEnvironment.get_req_id(
                    self.step_deps,
                    self.source_deps,
                ),
                full_id="_default",
                arch=self._arch,
            )
        ]

    @property
    def env_id(self) -> EnvID:
        arch = self.requested_arch
        my_arch_env = [i for i in self.env_ids if i.arch == arch]
        if my_arch_env:
            return my_arch_env[0]

        raise InvalidEnvironmentException(
            "Architecture '%s' not requested for step" % arch
        )

    @property
    def non_base_source_deps(self) -> Sequence[TStr]:
        return self._resolve_deps_sources()[1]

    @property
    def non_base_step_deps(self) -> Sequence[TStr]:
        return self._resolve_deps_sources()[0]

    @property
    def source_deps(self) -> Sequence[TStr]:
        return self._resolve_deps_sources()[3]

    @property
    def step_deps(self) -> Sequence[TStr]:
        return self._resolve_deps_sources()[2]

    @property
    def env_type(self) -> Optional[EnvType]:
        # We return an env-type in only one case: this is an environment that has
        # a base environment and the derived environment is of the same type (PIP)
        self._resolve_deps_sources()
        return self._env_type

    @property
    def requested_arch(self) -> str:
        return self._arch

    @property
    def local_root(self) -> Optional[str]:
        return self._local_root

    @property
    def from_env_name(self) -> Optional[str]:
        return self._from()

    @property
    def from_env_name_unresolved(self) -> Optional[str]:
        return self._from(True)

    @property
    def from_env(self) -> Optional[ResolvedEnvironment]:
        from_alias = self._from()
        if from_alias is not None:
            if self._from_env:
                return self._from_env
            # Else, we need to resolve it
            self._get_conda(self._echo, self._flow_datastore_type)
            assert self.conda
            from_env_id = self.conda.env_id_from_alias(from_alias, self._arch)
            if not from_env_id:
                raise InvalidEnvironmentException(
                    "'%s' does not refer to a known Conda environment" % from_alias
                )
            # Here we have a valid env_id so we can now get it for this architecture
            self._from_env = self.conda.environment(from_env_id)
            if self._from_env is None:
                raise InvalidEnvironmentException(
                    "'%s' is a valid Conda environment but does not exist for %s"
                    % (from_alias, self._arch)
                )
        return self._from_env

    def set_conda(self, conda: Conda):
        self.conda = conda

    @staticmethod
    def sub_envvars_in_envname(
        name: str, addl_env: Optional[Dict[str, Union[str, Callable[[], str]]]] = None
    ) -> str:
        init_name = name
        if addl_env is None:
            addl_env = {}
        envvars_to_sub = re.findall(r"\@{(\w+)}", name)
        for envvar in set(envvars_to_sub):
            replacement = os.environ.get(envvar, addl_env.get(envvar))
            if callable(replacement):
                replacement = replacement()
            if replacement is not None:
                name = name.replace("@{%s}" % envvar, replacement)
            else:
                raise InvalidEnvironmentException(
                    "Could not find '%s' in the environment -- needed to resolve '%s'"
                    % (envvar, name)
                )
        return name

    def step_init(
        self,
        flow: FlowSpec,
        graph: FlowGraph,
        step_name: str,
        decorators: List[StepDecorator],
        environment: MetaflowEnvironment,
        flow_datastore: FlowDataStore,
        logger: Callable[..., None],
    ):
        if environment.TYPE != "conda":
            raise InvalidEnvironmentException(
                "The *@%s* decorator requires " "--environment=conda" % self.name
            )

        if not self._resolve_pip_or_conda_deco(flow, decorators):
            return
        self._echo = logger
        self._env = environment
        self._flow = flow
        self._step_name = step_name
        self._flow_datastore_type = flow_datastore.TYPE  # type: str
        self._flow_datastore = flow_datastore
        self._base_attributes = self._get_base_attributes()

        self._is_remote = any(
            [deco.name in CONDA_REMOTE_COMMANDS for deco in decorators]
        )
        if self._is_remote:
            self._arch = "linux-64"
        else:
            self._arch = arch_id()

        self.__class__._local_root = LocalStorage.get_datastore_root_from_config(
            self._echo
        )  # type: str

        # Information about the environment this environment is built from
        self._from_env = None  # type: Optional[ResolvedEnvironment]
        self._resolved_non_base_deps = None  # type: Optional[Sequence[TStr]]
        self._resolved_non_base_sources = None  # type: Optional[Sequence[TStr]]
        self._resolved_deps = None  # type: Optional[Sequence[TStr]]
        self._resolved_sources = None  # type: Optional[Sequence[TStr]]
        self._env_type = None  # type: Optional[EnvType]
        self._env_for_fetch = {}  # type: Dict[str, Union[str, Callable[[], str]]]
        self._flow = None  # type: Optional[FlowSpec]

        if (self.attributes["name"] or self.attributes["pathspec"]) and any(
            [
                True
                for k, v in self.attributes.items()
                if v and k not in ("name", "pathspec", "fetch_at_exec")
            ]
        ):
            raise InvalidEnvironmentException(
                "You cannot specify `name` or `pathspec` along with other attributes in @%s"
                % self.name
            )

        if self.is_fetch_at_exec():
            if not self._from(raw_name=True):
                raise InvalidEnvironmentException(
                    "You cannot specify a `fetch_at_exec` environment and no environment "
                    "to fetch (either through `name` or `pathspec`) in @%s" % self.name
                )
            # We are also very strict that the environment should be *only* a name
            # and nothing else as we won't re-resolve
            if any(
                [
                    True
                    for k, v in self.attributes.items()
                    if v and k not in ("name", "pathspec", "fetch_at_exec")
                ]
            ) or any(
                [
                    True
                    for k, v in self._base_attributes.items()
                    if v and k not in ("name", "pathspec", "fetch_at_exec")
                ]
            ):
                raise InvalidEnvironmentException(
                    "You cannot specify a `fetch_at_exec` environment with anything "
                    "other than a pure named environment in @%s" % self.name
                )

        os.environ["PYTHONNOUSERSITE"] = "1"

    def runtime_init(self, flow: FlowSpec, graph: FlowGraph, package: Any, run_id: str):
        # Create a symlink to installed version of metaflow to execute user code against
        path_to_metaflow = os.path.join(get_metaflow_root(), "metaflow")
        path_to_info = os.path.join(get_metaflow_root(), "INFO")
        self._metaflow_home = tempfile.mkdtemp(dir="/tmp")
        self._addl_paths = None
        os.symlink(path_to_metaflow, os.path.join(self._metaflow_home, "metaflow"))

        # Symlink the INFO file as well to properly propagate down the Metaflow version
        # if launching on AWS Batch for example
        if os.path.isfile(path_to_info):
            os.symlink(path_to_info, os.path.join(self._metaflow_home, "INFO"))
        else:
            # If there is no "INFO" file, we will actually create one in this new
            # place because we won't be able to properly resolve the EXT_PKG extensions
            # the same way as outside conda (looking at distributions, etc). In a
            # Conda environment, as shown below (where we set self._addl_paths), all
            # EXT_PKG extensions are PYTHONPATH extensions. Instead of re-resolving,
            # we use the resolved information that is written out to the INFO file.
            with open(
                os.path.join(self._metaflow_home, "INFO"), mode="wt", encoding="utf-8"
            ) as f:
                f.write(
                    json.dumps(self._env.get_environment_info(include_ext_info=True))
                )

        # Do the same for EXT_PKG
        try:
            m = importlib.import_module(EXT_PKG)
        except ImportError:
            # No additional check needed because if we are here, we already checked
            # for other issues when loading at the toplevel
            pass
        else:
            custom_paths = list(set(m.__path__))  # For some reason, at times, unique
            # paths appear multiple times. We simplify
            # to avoid un-necessary links

            if len(custom_paths) == 1:
                # Regular package; we take a quick shortcut here
                os.symlink(
                    custom_paths[0],
                    os.path.join(self._metaflow_home, EXT_PKG),
                )
            else:
                # This is a namespace package, we therefore create a bunch of directories
                # so we can symlink in those separately and we will add those paths
                # to the PYTHONPATH for the interpreter. Note that we don't symlink
                # to the parent of the package because that could end up including
                # more stuff we don't want
                self._addl_paths = []  # type: List[str]
                for p in custom_paths:
                    temp_dir = tempfile.mkdtemp(dir=self._metaflow_home)
                    os.symlink(p, os.path.join(temp_dir, EXT_PKG))
                    self._addl_paths.append(temp_dir)

        # Also install any environment escape overrides directly here to enable
        # the escape to work even in non metaflow-created subprocesses
        generate_trampolines(self._metaflow_home)

        # If we need to fetch the environment on exec, save the information we need
        # so that we can resolve it using information such as run id, step name, task
        # id and parameter values
        if self.is_enabled() and self.is_fetch_at_exec():
            self._flow = flow
            self._env_for_fetch["METAFLOW_RUN_ID"] = run_id
            self._env_for_fetch["METAFLOW_STEP_NAME"] = self.name

    def runtime_task_created(
        self,
        task_datastore: TaskDataStore,
        task_id: str,
        split_index: int,
        input_paths: List[str],
        is_cloned: bool,
        ubf_context: str,
    ):
        if self.is_enabled(ubf_context) and self.is_fetch_at_exec(ubf_context):
            # We need to ensure we can properly find the environment we are
            # going to run in
            run_id, step_name, task_id = input_paths[0].split("/")
            parent_ds = self._flow_datastore.get_task_datastore(
                run_id, step_name, task_id
            )
            for var, _ in self._flow._get_parameters():
                self._env_for_fetch[
                    "METAFLOW_INIT_%s" % var.upper().replace("-", "_")
                ] = lambda _param=getattr(
                    self._flow, var
                ), _var=var, _ds=parent_ds: str(
                    _param.load_parameter(_ds[_var])
                )
            self._env_for_fetch["METAFLOW_TASK_ID"] = task_id

            self._get_conda(self._echo, self._flow_datastore_type)
            assert self.conda
            # Calling from_env_name will resolve the environment name using all the
            # additional variables injected above.
            resolved_env_id = self.conda.env_id_from_alias(
                cast(str, self.from_env_name), arch=self._arch
            )
            if resolved_env_id is None:
                raise RuntimeError(
                    "Cannot find environment '%s' (from '%s') for arch '%s'"
                    % (self.from_env_name, self.from_env_name_unresolved, self._arch)
                )

    def runtime_step_cli(
        self,
        cli_args: Any,  # Importing CLIArgs causes an issue so ignore for now
        retry_count: int,
        max_user_code_retries: int,
        ubf_context: str,
    ):
        # We also set the env var in remote case for is_fetch_at_exec
        # so that it can be used to fill out the bootstrap command with
        # the proper environment
        if self.is_enabled(UBF_TASK) or self.is_fetch_at_exec(ubf_context):
            self._get_conda(self._echo, self._flow_datastore_type)
            assert self.conda
            resolved_env = cast(
                ResolvedEnvironment, self.conda.environment(self.env_id)
            )
            my_env_id = resolved_env.env_id
            # Export this for local runs, we will use it to read the "resolved"
            # environment ID in task_pre_step; this makes it compatible with the remote
            # bootstrap which also exports it. We do this even for UBF control tasks as
            # this environment variable is then passed to the actual tasks. We don't create
            # the environment for the control task -- just for the actual tasks.
            cli_args.env["_METAFLOW_CONDA_ENV"] = json.dumps(my_env_id)

        if not self.is_enabled(ubf_context) or self._is_remote:
            return
        # Create the environment we are going to use

        if self.conda.created_environment(my_env_id):
            self._echo(
                "Using existing Conda environment %s (%s)"
                % (my_env_id.req_id, my_env_id.full_id)
            )
        else:
            # Otherwise, we read the conda file and create the environment locally
            self._echo(
                "Creating Conda environment %s (%s)..."
                % (my_env_id.req_id, my_env_id.full_id)
            )
            self.conda.create_for_step(self._step_name, resolved_env)

        # Actually set it up.
        python_path = self._metaflow_home
        if self._addl_paths is not None:
            addl_paths = os.pathsep.join(self._addl_paths)
            python_path = os.pathsep.join([addl_paths, python_path])

        cli_args.env["PYTHONPATH"] = python_path
        entrypoint = self.conda.python(my_env_id)
        if entrypoint is None:
            # This should never happen -- it means the environment was not
            # created somehow
            raise InvalidEnvironmentException("No executable found for environment")
        cli_args.entrypoint[0] = entrypoint

    def task_pre_step(
        self,
        step_name: str,
        task_datastore: TaskDataStore,
        metadata: MetadataProvider,
        run_id: str,
        task_id: str,
        flow: FlowSpec,
        graph: FlowGraph,
        retry_count: int,
        max_user_code_retries: int,
        ubf_context: str,
        inputs: List[str],
    ):
        if self.is_enabled(ubf_context):
            # Add the Python interpreter's parent to the path. This is to
            # ensure that any non-pythonic dependencies introduced by the conda
            # environment are visible to the user code.
            env_path = os.path.dirname(os.path.realpath(sys.executable))
            if os.environ.get("PATH") is not None:
                env_path = os.pathsep.join([env_path, os.environ["PATH"]])
            os.environ["PATH"] = env_path

            metadata.register_metadata(
                run_id,
                step_name,
                task_id,
                [
                    MetaDatum(
                        field="conda_env_id",
                        value=os.environ["_METAFLOW_CONDA_ENV"],
                        type="conda_env_id",
                        tags=["attempt_id:{0}".format(retry_count)],
                    )
                ],
            )

    def runtime_finished(self, exception: Exception):
        shutil.rmtree(self._metaflow_home)

    def _get_base_attributes(self) -> Dict[str, Any]:
        if "pip_base" in self._flow._flow_decorators:
            raise InvalidEnvironmentException(
                "@conda decorator is not compatible with @pip_base decorator."
            )
        if "conda_base" in self._flow._flow_decorators:
            return self._flow._flow_decorators["conda_base"][0].attributes
        return self.defaults

    def _self_disabled(self) -> bool:
        self_disabled = self.attributes["disabled"]
        if self_disabled is None:
            # If the user sets anything we consider that disabled = False
            if (
                self.attributes["name"]
                or self.attributes["pathspec"]
                or self.attributes["libraries"]
                or self.attributes["pip_packages"]
                or self.attributes["python"]
            ):
                self_disabled = False
        return self_disabled

    def _python_version(self) -> str:
        return next(
            x
            for x in [
                self.attributes["python"],
                self._base_attributes["python"],
                self._from_env_python(),
                platform.python_version(),
            ]
            if x is not None
        )

    def _from(self, raw_name: bool = False) -> Optional[str]:
        possible_name = (
            next(
                x
                for x in [
                    self.attributes["name"],
                    "step:%s" % self.attributes["pathspec"]
                    if self.attributes["pathspec"]
                    else None,
                    self._base_attributes["name"],
                    "step:%s" % self._base_attributes["pathspec"]
                    if self._base_attributes["pathspec"]
                    else None,
                    "",
                ]
                if x is not None
            )
            or None
        )

        if possible_name is None:
            return None

        possible_name = cast(str, possible_name)
        if raw_name:
            return possible_name

        # Substitute environment variables
        return self.sub_envvars_in_envname(possible_name, self._env_for_fetch)

    def _from_env_python(self) -> Optional[str]:
        from_env = self.from_env
        if from_env:
            for p in from_env.packages:
                if p.package_name == "python":
                    return p.package_version
            raise InvalidEnvironmentException(
                "Cannot determine Python version from the base environment"
            )
        return None

    def _np_conda_deps(self) -> Dict[str, str]:
        return {}

    def _conda_deps(self) -> Dict[str, str]:
        deps = {}  # type: Dict[str, str]
        if self.from_env:
            if self.from_env.env_type != EnvType.PIP_ONLY:
                # We don't get pinned deps here -- we will set them as pip ones to
                # allow things like @conda_base(name=<piponlyenv>)
                deps = dict(
                    get_pinned_conda_libs(
                        self._python_version(), self._flow_datastore_type
                    )
                )
        else:
            deps = dict(
                get_pinned_conda_libs(self._python_version(), self._flow_datastore_type)
            )

        # Things in the @conda decorator replace the ones in the base decorator
        user_deps = dict(self._base_attributes["libraries"])
        user_deps.update(self.attributes["libraries"])

        # We merge with the base deps so user can't override what we need
        return merge_dep_dicts(deps, user_deps)

    def _conda_channels(self) -> List[str]:
        seen = set()  # type: Set[str]
        result = []  # type: List[str]
        for c in chain(
            self.attributes["channels"],
            self._base_attributes["channels"],
        ):
            if c in seen:
                continue
            seen.add(c)
            result.append(c)
        return result

    def _pip_deps(self) -> Dict[str, str]:
        deps = {}  # type: Dict[str, str]
        if self.from_env and self.from_env == EnvType.PIP_ONLY:
            deps = conda_deps_to_pip_deps(
                get_pinned_conda_libs(self._python_version(), self._flow_datastore_type)
            )

        user_deps = dict(self._base_attributes["pip_packages"])
        user_deps.update(self.attributes["pip_packages"])

        return merge_dep_dicts(deps, user_deps)

    def _pip_sources(self) -> List[str]:
        seen = set()  # type: Set[str]
        result = []  # type: List[str]
        for c in chain(
            self.attributes["pip_sources"],
            self._base_attributes["pip_sources"],
        ):
            if c in seen:
                continue
            seen.add(c)
            result.append(c)
        return result

    def _resolve_deps_sources(
        self,
    ) -> Tuple[Sequence[TStr], Sequence[TStr], Sequence[TStr], Sequence[TStr]]:
        if all(
            [
                self._resolved_non_base_deps,
                self._resolved_non_base_sources,
                self._resolved_deps,
                self._resolved_sources,
            ]
        ):
            return (
                self._resolved_non_base_deps,
                self._resolved_non_base_sources,
                self._resolved_deps,
                self._resolved_sources,
            )

        self._resolved_non_base_deps = []
        # Empty version will just be "I want this package with no version constraints"
        self._resolved_non_base_deps.extend(
            TStr("conda", "%s==%s" % (name, ver) if ver else name)
            for name, ver in self._conda_deps().items()
        )

        # We keep the same env-type if we can.
        self._env_type = (
            EnvType.PIP_ONLY
            if self.from_env
            and self.from_env.env_type == EnvType.PIP_ONLY
            and len(self._resolved_non_base_deps) == 0
            else None
        )

        self._resolved_non_base_deps.extend(
            TStr("npconda", "%s==%s" % (name, ver) if ver else name)
            for name, ver in self._np_conda_deps().items()
        )
        self._resolved_non_base_deps.extend(
            TStr(
                "pip",
                "%s==%s" % (name, canonicalize_version(ver)) if ver else name,
            )
            for name, ver in self._pip_deps().items()
        )
        if not self.from_env:
            self._resolved_non_base_deps.append(
                TStr(
                    "conda",
                    "python==%s" % canonicalize_version(self._python_version()),
                )
            )

        self._resolved_non_base_sources = []
        self._resolved_non_base_sources.extend(
            map(
                lambda x: TStr("conda", x),
                self._conda_channels(),
            )
        )
        self._resolved_non_base_sources.extend(
            map(
                lambda x: TStr("pip", x),
                self._pip_sources(),
            )
        )

        if self.from_env:
            from .envsresolver import EnvsResolver  # Avoid circular import

            # We need to recompute the req ID based on the base environment
            self._get_conda(self._echo, self._flow_datastore_type)
            assert self.conda
            # Maybe we can get rid of this. Not sure.
            (
                _,
                self._resolved_sources,
                self._resolved_deps,
                _,
                _,
            ) = EnvsResolver.extract_info_from_base(
                self.conda,
                self.from_env,
                self._resolved_non_base_deps,
                self._resolved_non_base_sources,
                [],
                self.from_env.env_id.arch,
            )
        else:
            self._resolved_deps = self._resolved_non_base_deps
            self._resolved_sources = self._resolved_non_base_sources
        return (
            self._resolved_non_base_deps,
            self._resolved_non_base_sources,
            self._resolved_deps,
            self._resolved_sources,
        )

    def _resolve_pip_or_conda_deco(
        self, flow: FlowSpec, decorators: List[StepDecorator]
    ) -> bool:
        has_pip_base = "pip_base" in flow._flow_decorators
        has_conda_base = "conda_base" in flow._flow_decorators

        # Note that other decorators *extend* either a conda decorator or a pip decorator
        # so we look for those decorators as well.
        # The pip decorator also extends the conda decorator but here we mean more
        # in terms of functionality:
        #  - extending a conda decorator means providing both pip and conda dependencies
        #    (potentially)
        #  - extending a pip decorator means providing only pip packages
        all_decs = [d for d in decorators if isinstance(d, CondaStepDecorator)]
        last_deco = all_decs[-1]
        conda_decs = [d for d in all_decs if d.TYPE == "conda"]
        pip_decs = [d for d in all_decs if d.TYPE == "pip"]

        to_remove = []
        if len(conda_decs) > 1:
            # There is at least one user defined decorator so we remove all the others
            to_remove = [x for x in conda_decs if not x.statically_defined]
            conda_decs = [x for x in conda_decs if x.statically_defined]
        if len(pip_decs) > 1:
            # Ditto for pip
            to_remove = [x for x in pip_decs if not x.statically_defined]
            pip_decs = [x for x in pip_decs if x.statically_defined]

        # In the environment with decospecs, we add both conda and pip
        # decorators so that we can choose the best one based on the presence of the
        # base decorators for example. In this function, we clean up the
        # decorators and remove all the extraneous ones. In some cases, however,
        # this function can be called multiple times (when deploying to a scheduler,
        # the step_init (which calls this) is called twice). In that case, we just
        # continue along as we already cleaned things out the first time.
        if len(conda_decs) == 0 or len(pip_decs) == 0:
            return True

        if len(conda_decs) > 1:
            # We can only have one Conda-type decorator
            raise InvalidEnvironmentException(
                "Multiple decorators (%s) provide @conda-like functionality for "
                "step '%s'. Please add only one such decorator and include any "
                "additional dependencies using the same arguments as for @conda."
                % (", ".join(["@%s" % d.name for d in conda_decs]), self.name)
            )
        if len(pip_decs) > 1:
            # Ditto with Pip-type decorators
            raise InvalidEnvironmentException(
                "Multiple decorators (%s) provide @pip-like functionality for "
                "step '%s'. Please add only one such decorator and include any "
                "additional dependencies using the same arguments as for @pip."
                % (", ".join(["@%s" % d.name for d in pip_decs]), self.name)
            )

        conda_deco = conda_decs[0]
        pip_deco = pip_decs[0]

        debug.conda_exec(
            "In %s decorator: pip_base(%s), conda_base(%s), conda_deco(%s), pip_deco(%s)"
            % (self.name, has_pip_base, has_conda_base, conda_deco.name, pip_deco.name)
        )
        if conda_deco.statically_defined and pip_deco.statically_defined:
            raise InvalidEnvironmentException(
                "Cannot specify both @%s (Conda decorator) and @%s (Pip decorator) "
                "in step '%s'. If you need both pip and conda dependencies, "
                "use @%s and pass in the pip dependencies as `pip_packages` and "
                "the sources as `pip_sources`"
                % (
                    conda_deco.name,
                    pip_deco.name,
                    self.name,
                    conda_deco.name,
                )
            )
        if has_pip_base and conda_deco.statically_defined:
            raise InvalidEnvironmentException(
                "@pip_base is not compatible with @%s. "
                "Use @conda_base instead (using `pip_packages` instead of `packages` "
                "and `pip_sources` instead of `sources`)" % conda_deco.name
            )
        if has_conda_base and pip_deco.statically_defined:
            raise InvalidEnvironmentException(
                "@conda_base is not compatible with @%s. Use @pip_base if only using "
                "pip dependencies or replace @%s with a @conda decorator (using "
                "`pip_packages` instead of `packages` and `pip_sources` instead of "
                "`sources)" % (pip_deco.name, pip_deco.name)
            )

        # At this point, we have at most one statically defined so we keep that one
        # or the one derived from the base decorator.
        # If we have none, we keep the conda one (base one).
        # Return true if we should continue the function. False if we return (ie:
        # we are going to be deleted)
        del_deco = pip_deco
        if pip_deco.statically_defined or has_pip_base:
            del_deco = conda_deco
        to_remove.append(del_deco)
        # We remove only when we are the last decorator since this is called while
        # we are iterating on the list
        if self.name == last_deco.name:
            for d in to_remove:
                decorators.remove(d)
        return self.name not in [d.name for d in to_remove]

    @classmethod
    def _get_conda(cls, echo: Callable[..., None], datastore_type: str) -> None:
        if cls.conda is None:
            cls.conda = Conda(echo, datastore_type)


class PipStepDecorator(CondaStepDecorator):
    """
    Specifies the Pip environment for the step.

    Information in this decorator will augment any
    attributes set in the `@pip_base` flow-level decorator. Hence
    you can use `@pip_base` to set common libraries required by all
    steps and use `@pip` to specify step-specific additions.
    Information specified in this decorator will augment the information in the base
    decorator and, in case of a conflict (for example the same library specified in
    both the base decorator and the step decorator), the step decorator's information
    will prevail.

    Parameters
    ----------
    name : Optional[str]
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this step. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : Optional[str]
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    packages : Optional[Dict[str, str]]
        Packages to use for this step. The key is the name of the package
        and the value is the version to use (default: `{}`).
    sources : Optional[List[str]]
        Additional channels to search for
    python : Optional[str]
        Version of Python to use, e.g. '3.7.4'. If not specified, the current python
        version will be used.
    fetch_at_exec : bool, default False
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this step to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    disabled : bool, default False
        If set to True, disables Pip.
    """

    name = "pip"
    TYPE = "pip"

    defaults = {
        "name": None,
        "pathspec": None,
        "packages": {},
        "sources": [],
        "python": None,
        "fetch_at_exec": None,
        "disabled": None,
    }

    def _self_disabled(self) -> bool:
        self_disabled = self.attributes["disabled"]
        if self_disabled is None:
            # If the user sets anything we consider that disabled = False
            if (
                self.attributes["name"]
                or self.attributes["pathspec"]
                or self.attributes["packages"]
                or self.attributes["python"]
            ):
                self_disabled = False
        return self_disabled

    def _np_conda_deps(self) -> Dict[str, str]:
        return {}

    def _conda_deps(self) -> Dict[str, str]:
        if self.from_env and self.from_env.env_type != EnvType.PIP_ONLY:
            return dict(
                get_pinned_conda_libs(self._python_version(), self._flow_datastore_type)
            )
        return {}

    def _conda_channels(self) -> List[str]:
        return []

    def _pip_deps(self) -> Dict[str, str]:
        deps = {}  # type: Dict[str, str]
        if self.from_env:
            if self.from_env.env_type == EnvType.PIP_ONLY:
                deps = conda_deps_to_pip_deps(
                    get_pinned_conda_libs(
                        self._python_version(), self._flow_datastore_type
                    )
                )
        else:
            deps = conda_deps_to_pip_deps(
                get_pinned_conda_libs(self._python_version(), self._flow_datastore_type)
            )

        user_deps = dict(self._base_attributes["packages"])
        user_deps.update(self.attributes["packages"])

        return merge_dep_dicts(deps, user_deps)

    def _pip_sources(self) -> List[str]:
        seen = set()  # type: Set[str]
        result = []  # type: List[str]
        for c in chain(
            self.attributes["sources"],
            self._base_attributes["sources"],
        ):
            if c in seen:
                continue
            seen.add(c)
            result.append(c)
        return result

    def _get_base_attributes(self) -> Dict[str, Any]:
        if "conda_base" in self._flow._flow_decorators:
            raise InvalidEnvironmentException(
                "@pip decorator is not compatible with @conda_base decorator."
            )
        if "pip_base" in self._flow._flow_decorators:
            return self._flow._flow_decorators["pip_base"][0].attributes
        return self.defaults


def get_conda_decorator(flow: FlowSpec, step_name: str) -> CondaStepDecorator:
    step = next(step for step in flow if step.name == step_name)

    decorator = next(
        deco for deco in step.decorators if isinstance(deco, CondaStepDecorator)
    )
    # Guaranteed to have a conda decorator because of env.decospecs()
    return decorator



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

import time

from itertools import chain
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import (
    cast,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
)

from metaflow.debug import debug
from metaflow.metaflow_config import CONDA_PREFERRED_FORMAT
from metaflow.metaflow_environment import InvalidEnvironmentException

from .env_descr import EnvID, EnvType, PackageSpecification, ResolvedEnvironment, TStr
from .conda import Conda
from .conda_step_decorator import CondaStepDecorator
from .utils import (
    AliasType,
    CondaException,
    channel_from_url,
    arch_id,
    merge_dep_dicts,
    plural_marker,
    resolve_env_alias,
    split_into_dict,
)


class EnvsResolver(object):
    def __init__(self, conda: Conda):
        # key: EnvID; value: dict containing:
        #  - "id": key
        #  - "steps": steps using this environment
        #  - "user_deps": array of requested dependencies
        #  - "deps": full array of dependencies -- typically user_deps except if base_env
        #  - "sources": full array of sources
        #  - "extras": additional arguments (typically used for PIP)
        #  - "conda_format": list of formats we want the Conda packages in or ["_any"] if no pref
        #  - "base": optional base environment this environment derives from
        #  - "base_accurate": T/F; True if the "deps" for the base environment are known
        #  - "resolved": The resolved environment
        #  - "already_resolved": T/F: True if we had a resolved environment prior to resolving
        #  - "force": T/F: True if we need to re-resolve
        self._requested_envs = {}  # type: Dict[EnvID, Dict[str, Any]]
        self._builder_envs = {}  # type: Dict[EnvID, Dict[str, Any]]
        self._conda = conda
        self._non_step_envs = False
        self._co_resolved_force_resolve = set()  # type: Set[str]

    def add_environment(
        self,
        architecture: str,
        user_deps: Sequence[TStr],
        user_sources: Sequence[TStr],
        extras: Sequence[TStr],
        base_env: Optional[ResolvedEnvironment] = None,
        env_type: Optional[EnvType] = None,
        base_from_full_id: bool = False,
        local_only: bool = False,
        force: bool = False,
        force_co_resolve: bool = False,
    ):
        """
        Add an environment to resolve to this EnvsResolver. The EnvsResolver will resolve
        the same environment only once and will mutualize anything it can.


        Parameters
        ----------
        architecture : str
            Architecture this environment is requested for
        user_deps : Sequence[TStr]
            The user dependencies of this environment -- if there is a base environment,
            these are the *additional* dependencies.
        user_sources : Sequence[TStr]
            The sources of this environment -- if there is a base environment, these are
            the *additional* sources.
        extras : Sequence[TStr]
            Any extra information to pass to Conda/Pip (currently just pip) -- if there
            is a base environment, these are the *additional* extras.
        base_env : Optional[ResolvedEnvironment], optional
            The base environment, if any, this environment is derived from, by default None
        env_type : Optional[EnvType], optional
            The environment type -- this is only used if base_env is not None and
            we know what the derived environment should be. Else it is automatically
            determined.
        base_from_full_id : bool, optional
            True if the base environment was extracted using just the full ID of the
            environment which makes it imprecise, by default False
        local_only : bool, optional
            True if we should only look for resolved environments locally and not
            remotely, by default False
        force : bool, optional
            True if we should force resolution even if this environment is known,
            by default False
        force_co_resolve : bool, optional
            True if we should force co-resolution, by default False
        """
        self._non_step_envs = True

        # If there is a base environment, get all the resolved information
        if base_env:
            (
                env_id,
                user_sources,
                user_deps,
                deps,
                extras,
            ) = self.extract_info_from_base(
                self._conda, base_env, user_deps, user_sources, extras, architecture
            )
        else:
            env_id = EnvID(
                ResolvedEnvironment.get_req_id(user_deps, user_sources, extras),
                "_default",
                architecture,
            )
            deps = user_deps

        if (
            not force
            and base_env
            and base_env.env_id.req_id == env_id.req_id
            and base_env.env_id.arch == env_id.arch
        ):
            resolved_env = base_env
        else:
            resolved_env = (
                self._conda.environment(env_id, local_only) if not force else None
            )
        if env_id not in self._requested_envs:
            if force_co_resolve:
                if resolved_env is None:
                    # Invalidate any previously resolved environment with the same req_id
                    for other_env_id, other_env in self._requested_envs.items():
                        if other_env_id.req_id == env_id.req_id:
                            other_env["resolved"] = None
                            other_env["already_resolved"] = False
                    self._co_resolved_force_resolve.add(env_id.req_id)
                elif env_id.req_id in self._co_resolved_force_resolve:
                    # If another environment for the same req-id is not resolved, we
                    # need to resolve this one too.
                    resolved_env = None

            self._requested_envs[env_id] = {
                "id": env_id,
                "steps": ["ad-hoc"],
                "user_deps": user_deps,
                "deps": deps,
                "sources": user_sources,
                "extras": extras,
                "conda_format": [CONDA_PREFERRED_FORMAT]
                if CONDA_PREFERRED_FORMAT
                else ["_any"],
                "base": base_env,
                "base_accurate": base_env
                and base_env.is_info_accurate
                and not base_from_full_id,
                "resolved": resolved_env,
                "already_resolved": resolved_env is not None,
                "env_type": env_type,
                "force": force,
            }
            debug.conda_exec(
                "Added environment to resolve %s" % str(self._requested_envs[env_id])
            )

    def add_environment_for_step(
        self,
        step_name: str,
        decorator: CondaStepDecorator,
        force: bool = False,
        archs: Optional[List[str]] = None,
    ):
        """
        This is the same as `add_environment` except a lot of the information needed
        for `add_environment` is extracted from the step decorator passed in

        Parameters
        ----------
        step_name : str
            Step name for which this environment is being resolved (used for debugging
            messages)
        decorator : CondaStepDecorator
            Decorator to use to extract information from
        force : bool, optional
            True if we should resolve the environment even if we already
            know about it, by default False
        archs : Optional[List[str]], optional
            List of architectures to resolve for. If None, resolves for
            the current architecture, by default None
        """
        # Make sure we use the same conda object to properly cache everything
        decorator.set_conda(self._conda)

        from_env = decorator.from_env
        if from_env:
            (
                env_id,
                user_sources,
                user_deps,
                deps,
                extras,
            ) = self.extract_info_from_base(
                self._conda,
                from_env,
                decorator.non_base_step_deps,
                decorator.source_deps,
                [],
                arch_id(),
            )
            env_ids = [env_id]
        else:
            user_deps = decorator.non_base_step_deps
            user_sources = decorator.source_deps
            deps = decorator.step_deps

        env_ids = decorator.env_ids
        base_req_id = env_ids[0].req_id
        base_full_id = env_ids[0].full_id
        env_ids = [
            EnvID(req_id=base_req_id, full_id=base_full_id, arch=arch)
            for arch in set((archs or []) + [env_id.arch for env_id in env_ids])
        ]

        for env_id in env_ids:
            if (
                not force
                and from_env
                and env_id.req_id == from_env.env_id.req_id
                and env_id.arch == from_env.env_id.arch
            ):
                debug.conda_exec(
                    "For step '%s', found a clean base environment -- using that"
                    % step_name
                )
                resolved_env = from_env
            else:
                resolved_env = self._conda.environment(env_id) if not force else None
            if env_id not in self._requested_envs:
                self._requested_envs[env_id] = {
                    "id": env_id,
                    "steps": [step_name],
                    "user_deps": user_deps,
                    "deps": deps,
                    "sources": user_sources,
                    "extras": [],
                    "conda_format": [CONDA_PREFERRED_FORMAT]
                    if CONDA_PREFERRED_FORMAT
                    else ["_any"],
                    "base": from_env,
                    "base_accurate": from_env
                    and from_env.is_info_accurate
                    and resolve_env_alias(cast(str, decorator.from_env_name))[0]
                    != AliasType.FULL_ID,
                    "resolved": resolved_env,
                    "already_resolved": resolved_env is not None,
                    "env_type": decorator.env_type,
                    "force": force,
                }
                debug.conda_exec(
                    "Added environment to resolve %s"
                    % str(self._requested_envs[env_id])
                )
            else:
                self._requested_envs[env_id]["steps"].append(step_name)
                debug.conda_exec(
                    "Environment '%s' is also needed by '%s'" % (env_id, step_name)
                )

    def resolve_environments(self, echo: Callable[..., None]):
        """
        Resolve the environments added using `add_environment` and `add_environment_for_step`.

        Parameters
        ----------
        echo : Callable[..., None]
            Method to use to print things to the console
        """
        # At this point, we check in our backend storage if we have the files we need
        need_resolution = [
            env_id
            for env_id, req in self._requested_envs.items()
            if req["resolved"] is None
        ]
        if debug.conda:
            debug.conda_exec("Resolving environments:")
            for env_id in need_resolution:
                info = self._requested_envs[env_id]
                debug.conda_exec(
                    "%s (%s): %s" % (env_id.req_id, env_id.full_id, str(info))
                )
        if len(need_resolution):
            self._resolve_environments(echo, need_resolution)

    def all_environments(
        self, include_builder_envs: bool = False
    ) -> Iterator[Tuple[EnvID, Optional[ResolvedEnvironment], List[str]]]:
        """
        Returns an iterator over all the environments (resolved or not) this
        EnvsResolver knows about

        Parameters
        ----------
        include_builder_envs : bool, optional
            If True, will also return builder environments that were built
            while resolving the main environments. In the case of PIP environments,
            we sometimes need to build another environment to build PIP packages
            and/or resolve the PIP dependencies, by default False

        Yields
        ------
        Iterator[Tuple[EnvID, Optional[ResolvedEnvironment], List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the ResolvedEnvironment (if it was resolved)
                - the list of step names this environment pertains to.
        """
        src_dicts = [self._requested_envs]
        if include_builder_envs:
            src_dicts.append(self._builder_envs)
        for d in src_dicts:
            for env_id, req in d.items():
                yield env_id, cast(
                    Optional[ResolvedEnvironment], req["resolved"]
                ), cast(List[str], req["steps"])

    def resolved_environments(
        self, include_builder_envs: bool = False
    ) -> Iterator[Tuple[EnvID, ResolvedEnvironment, List[str]]]:
        """
        Same as `all_environments` except for resolved environments only

        Parameters
        ----------
        include_builder_envs : bool, optional
            If True, will also return builder environments that were built
            while resolving the main environments. In the case of PIP environments,
            we sometimes need to build another environment to build PIP packages
            and/or resolve the PIP dependencies, by default False

        Yields
        ------
        Iterator[Tuple[EnvID, ResolvedEnvironment, List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the ResolvedEnvironment
                - the list of step names this environment pertains to.
        """
        src_dicts = [self._requested_envs]
        if include_builder_envs:
            src_dicts.append(self._builder_envs)
        for d in src_dicts:
            for env_id, req in d.items():
                if req["resolved"] is not None:
                    yield env_id, cast(ResolvedEnvironment, req["resolved"]), cast(
                        List[str], req["steps"]
                    )

    def non_resolved_environments(
        self,
    ) -> Iterator[Tuple[EnvID, List[str]]]:
        """
        The opposite of `resolved_environments`. It never includes builder environments
        as those are always resolved

        Yields
        ------
        Iterator[Tuple[EnvID, List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the list of step names this environment pertains to.
        """
        for env_id, req in self._requested_envs.items():
            if req["resolved"] is None:
                yield env_id, cast(List[str], req["steps"])

    def need_caching_environments(
        self, include_builder_envs: bool = False
    ) -> Iterator[Tuple[EnvID, ResolvedEnvironment, List[str], List[str]]]:
        """
        Same as `resolved_environments` but further limits the output to environments
        that need to be cached.

        Parameters
        ----------
        include_builder_envs : bool, optional
            If True, will also return builder environments that were built
            while resolving the main environments. In the case of PIP environments,
            we sometimes need to build another environment to build PIP packages
            and/or resolve the PIP dependencies, by default False

        Yields
        ------
        Iterator[Tuple[EnvID, ResolvedEnvironment, List[str], List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the ResolvedEnvironment
                - the Conda formats that need caching
                - the list of step names this environment pertains to.
        """
        src_dicts = [self._requested_envs]
        if include_builder_envs:
            src_dicts.append(self._builder_envs)
        for d in src_dicts:
            for env_id, req in d.items():
                # Resolved environments that were not already resolved or not fully
                # cached
                if req["resolved"] is not None and (
                    not req["already_resolved"]
                    or not req["resolved"].is_cached({"conda": req["conda_format"]})
                ):
                    yield env_id, cast(ResolvedEnvironment, req["resolved"]), cast(
                        List[str], req["conda_format"]
                    ), cast(List[str], req["steps"])

    def new_environments(
        self, include_builder_envs: bool = False
    ) -> Iterator[Tuple[EnvID, ResolvedEnvironment, List[str]]]:
        """
        Same as `resolved_environments` but further limits the output to environments
        that were newly resolved (ie: excludes ones that were already resolved prior
        to starting this EnvsResolver)

        Parameters
        ----------
        include_builder_envs : bool, optional
            If True, will also return builder environments that were built
            while resolving the main environments. In the case of PIP environments,
            we sometimes need to build another environment to build PIP packages
            and/or resolve the PIP dependencies, by default False

        Yields
        ------
        Iterator[Tuple[EnvID, ResolvedEnvironment, List[str]]]
            Each tuple element contains:
                - the environment ID for the environment
                - the ResolvedEnvironment
                - the list of step names this environment pertains to.
        """
        src_dicts = [self._requested_envs]
        if include_builder_envs:
            src_dicts.append(self._builder_envs)
        for d in src_dicts:
            for env_id, req in d.items():
                # Resolved environments that were not already resolved
                if req["resolved"] is not None and not req["already_resolved"]:
                    yield env_id, cast(ResolvedEnvironment, req["resolved"]), cast(
                        List[str], req["steps"]
                    )

    def _resolve_environments(
        self, echo: Callable[..., None], env_ids: Sequence[EnvID]
    ):
        start = time.time()
        if len(env_ids) == len(self._requested_envs):
            echo(
                "    Resolving %d environment%s %s..."
                % (
                    len(env_ids),
                    plural_marker(len(env_ids)),
                    "in flow " if not self._non_step_envs else "",
                ),
                nl=False,
            )
        else:
            echo(
                "    Resolving %d of %d environment%s %s(others are cached) ..."
                % (
                    len(env_ids),
                    len(self._requested_envs),
                    plural_marker(len(self._requested_envs)),
                    "in flow " if not self._non_step_envs else "",
                ),
                nl=False,
            )

        def _resolve(
            env_desc: Mapping[str, Any],
            builder_environments: Optional[Dict[str, EnvID]],
        ) -> Tuple[EnvID, ResolvedEnvironment, Optional[ResolvedEnvironment]]:
            env_id = cast(EnvID, env_desc["id"])
            builder_env_id = (
                builder_environments.get(env_id.req_id, None)
                if builder_environments
                else None
            )
            builder_env = cast(
                Optional[ResolvedEnvironment],
                (
                    self._builder_envs[builder_env_id]["resolved"]
                    if builder_env_id
                    else None
                ),
            )
            resolved_env, builder_env = self._conda.resolve(
                env_desc["steps"],
                env_desc["deps"],
                env_desc["sources"],
                env_desc["extras"],
                env_id.arch,
                env_desc.get("env_type"),
                builder_env=builder_env,
                base_env=env_desc["base"],
            )
            if env_desc["base"]:
                # We try to copy things over from the base environment as it contains
                # potential caching information we don't need to rebuild. This also
                # properly sets the user dependencies to what we want as opposed to
                # including everything we resolved for.
                merged_packages = []  # type: List[PackageSpecification]
                base_packages = {
                    p.filename: p.to_dict()
                    for p in cast(ResolvedEnvironment, env_desc["base"]).packages
                }
                for p in resolved_env.packages:
                    existing_info = base_packages.get(p.filename)
                    if existing_info:
                        merged_packages.append(
                            PackageSpecification.from_dict(existing_info)
                        )
                    else:
                        merged_packages.append(p)
                resolved_env = ResolvedEnvironment(
                    env_desc["user_deps"],
                    env_desc["sources"],
                    env_desc["extras"],
                    env_id.arch,
                    all_packages=merged_packages,
                    env_type=resolved_env.env_type,
                    accurate_source=env_desc["base_accurate"],
                )
            return env_id, resolved_env, builder_env

        # In PIP-ONLY scenarios, we actually need a sub-environment to properly resolve
        # the pip environment (due to https://github.com/pypa/pip/issues/11664).
        # We try to mutualize these environments.
        # There are other cases where we need a builder env but we only build it
        # for user in PIP-ONLY scenarios because we are not sure we need it in other
        # cases. If we do need it in the other cases, it will be lazily built and returned
        # so we can cache them for later use (they just won't be re-used and cached
        # this time so it is possible that we resolve the same builder environment
        # multiple times)

        builders_by_req_id = {}  # type: Dict[str, Any]
        my_arch = arch_id()
        for env_id in env_ids:
            env_type = self._requested_envs[env_id].get("env_type")
            if env_type is None:
                env_type = self._conda.env_type_for_deps(
                    self._requested_envs[env_id]["deps"]
                )
            if env_type != EnvType.PIP_ONLY:
                # If this is not a PIP_ONLY environment -- do not build builder_env
                continue

            if env_id.arch != my_arch:
                # If this is not for this architecture, we will either build a vanilla
                # builder env (later) or wait for the environment for this architecture
                # (if it is another env_id in env_ids)
                builders_by_req_id.setdefault(
                    env_id.req_id, self._requested_envs[env_id]
                )
                continue

            # Builder environments only have pip and npconda dependencies

            to_resolve_info = self._requested_envs[env_id]

            builder_user_deps = [
                d for d in to_resolve_info["user_deps"] if d.category == "npconda"
            ] + [
                d
                for d in to_resolve_info["deps"]
                if d.category == "conda" and d.value.startswith("python==")
            ]
            builder_sources = [
                s for s in to_resolve_info["sources"] if s.category == "conda"
            ]

            builder_env_id = EnvID(
                ResolvedEnvironment.get_req_id(builder_user_deps, builder_sources, []),
                "_default",
                env_id.arch,
            )
            if builder_env_id in self._builder_envs:
                builder_env_info = self._builder_envs[builder_env_id]
                builder_env_info["steps"].extend(to_resolve_info["steps"])
            else:
                builder_env = None
                if not to_resolve_info["force"]:
                    builder_env = self._conda.environment(builder_env_id)
                builder_env_info = {
                    "id": builder_env_id,
                    "steps": to_resolve_info["steps"],
                    "user_deps": builder_user_deps,
                    "deps": builder_user_deps,
                    "sources": builder_sources,
                    "extras": [],
                    "conda_format": to_resolve_info["conda_format"],
                    "base": None,
                    "base_accurate": False,
                    "resolved": builder_env,
                    "already_resolved": builder_env is not None,
                    "env_type": EnvType.CONDA_ONLY,
                }

            builders_by_req_id[env_id.req_id] = builder_env_id
            self._builder_envs[builder_env_id] = builder_env_info

        # We need to make sure we have a builder environment for this architecture
        # at least. If we can, we reuse the one that is requested but if not, we create
        # an empty one that simply has the python version we need and nothing else.
        for req_id, v in builders_by_req_id.items():
            if not isinstance(v, EnvID):
                base_info = v
                builder_deps = [
                    d for d in base_info["deps"] if d.value.startswith("python==")
                ]
                builder_sources = [
                    s for s in base_info["sources"] if s.category == "conda"
                ]
                builder_env_id = EnvID(
                    ResolvedEnvironment.get_req_id(
                        builder_deps,
                        builder_sources,
                        [],
                    ),
                    "_default",
                    my_arch,
                )
                if builder_env_id in self._builder_envs:
                    builder_env_info = self._builder_envs[builder_env_id]
                    builder_env_info["steps"].extend(base_info["steps"])
                else:
                    # Here we never force rebuild -- it is only to resolve PIP so don't
                    # really care too much
                    builder_env = self._conda.environment(builder_env_id)
                    builder_env_info = {
                        "id": builder_env_id,
                        "steps": base_info["steps"],
                        "user_deps": builder_deps,
                        "deps": builder_deps,
                        "sources": builder_sources,
                        "extras": [],
                        "conda_format": base_info["conda_format"],
                        "base": None,
                        "base_accurate": False,
                        "resolved": builder_env,
                        "already_resolved": builder_env is not None,
                        "env_type": EnvType.CONDA_ONLY,
                    }
                builders_by_req_id[req_id] = builder_env_id
                self._builder_envs[builder_env_id] = builder_env_info

        # Resolve all builder enviornments
        if len(self._builder_envs):
            debug.conda_exec("Builder environments: %s" % str(self._builder_envs))

            # Build the environments
            with ThreadPoolExecutor() as executor:
                resolution_result = [
                    executor.submit(_resolve, v, None)
                    for v in self._builder_envs.values()
                    if not v["already_resolved"]
                ]
                for f in as_completed(resolution_result):
                    env_id, resolved, _ = f.result()
                    self._builder_envs[env_id]["resolved"] = resolved

        # NOTE: Co-resolved environments allow you to resolve a bunch of "equivalent"
        # environments for different platforms. This is great as it can allow you to
        # run code on Linux and then instantiate an environment to look at it on Mac.
        # One issue though is that the set of packages on Linux may change while those
        # on mac may not (or vice versa) so it is possible to get in the following
        # situation:
        # - Co-resolve at time A:
        #   - Get linux full_id 123 and mac full_id 456
        # - Co-resolve later at time B:
        #   - Get linux full_id 123 and mac full_id 789
        # This is a problem because now the 1:1 correspondence between co-resolved
        # environments (important for figuring out which environment to use) is broken
        #
        # To solve this problem, we consider that co-resolved environments participate
        # in the computation of the full_id (basically a concatenation of all packages
        # across all co-resolved environments). This maintains the 1:1 correspondence.
        # It has a side benefit that we can use that same full_id for all co-resolved
        # environment making one easier to find from the other (instead of using indirect
        # links)
        co_resolved_envs = (
            {}
        )  # type: Dict[str, List[Tuple[EnvID, ResolvedEnvironment]]]
        if len(env_ids):
            with ThreadPoolExecutor() as executor:
                resolution_result = [
                    executor.submit(_resolve, v, builders_by_req_id)
                    for k, v in self._requested_envs.items()
                    if k in env_ids
                ]
                for f in as_completed(resolution_result):
                    env_id, resolved_env, addl_builder_env = f.result()
                    co_resolved_envs.setdefault(env_id.req_id, []).append(
                        (env_id, resolved_env)
                    )
                    if (
                        addl_builder_env
                        and addl_builder_env.env_id not in self._builder_envs
                    ):
                        # We "hack" the extract_from_base to get the proper user and
                        # full requirements from addl_builder_env
                        _, _, user_deps, full_deps, _ = self.extract_info_from_base(
                            self._conda,
                            addl_builder_env,
                            [],
                            [],
                            [],
                            addl_builder_env.env_id.arch,
                        )
                        self._builder_envs[addl_builder_env.env_id] = {
                            "id": addl_builder_env.env_id,
                            "steps": self._requested_envs[env_id]["steps"],
                            "user_deps": user_deps,
                            "deps": full_deps,
                            "sources": addl_builder_env.sources,
                            "extras": [],
                            "conda_format": self._requested_envs[env_id][
                                "conda_format"
                            ],
                            "base": None,
                            "base_accurate": False,
                            "resolved": addl_builder_env,
                            "already_resolved": False,
                            "env_type": EnvType.CONDA_ONLY,
                        }

            # Now we know all the co-resolved environments so we can compute the full
            # ID for all those environments
            for envs in co_resolved_envs.values():
                if len(envs) > 1:
                    ResolvedEnvironment.set_coresolved_full_id([x[1] for x in envs])

                for orig_env_id, resolved_env in envs:
                    resolved_env_id = resolved_env.env_id
                    cached_resolved_env = self._conda.environment(resolved_env_id)
                    # This checks if there is the same resolved environment already
                    # cached (in which case, we don't have to check a bunch of things
                    # so makes it nicer)
                    if cached_resolved_env:
                        resolved_env = cached_resolved_env
                        self._requested_envs[orig_env_id]["already_resolved"] = True

                    self._requested_envs[orig_env_id]["resolved"] = resolved_env
                    debug.conda_exec(
                        "For environment %s (%s) %s need packages %s"
                        % (
                            resolved_env_id.req_id,
                            resolved_env_id.full_id,
                            self._requested_envs[orig_env_id],
                            ", ".join(
                                sorted([p.filename for p in resolved_env.packages])
                            ),
                        )
                    )

        duration = int(time.time() - start)
        echo(" done in %d second%s." % (duration, plural_marker(duration)))

    @staticmethod
    def extract_info_from_base(
        conda: Conda,
        base_env: ResolvedEnvironment,
        deps: Sequence[TStr],
        sources: Sequence[TStr],
        extras: Sequence[TStr],
        architecture: str,
    ) -> Tuple[EnvID, Sequence[TStr], Sequence[TStr], Sequence[TStr], Sequence[TStr]]:
        """
        Returns the information about an environment extended from a given environment
        (in other words, taking the base environment first and then extending it
        with other dependencies).

        Parameters
        ----------
        base_env : ResolvedEnvironment
            The base environment to use
        deps : Sequence[TStr]
            The user dependencies to add/update
        sources : Sequence[TStr]
            The sources to add/update
        extras : Sequence[TStr]
            Additional information passed to Conda/Pip
        architecture : str
            The architecture to resolve for

        Returns
        -------
        Tuple[EnvID, Sequence[TStr], Sequence[TStr], Sequence[TStr], Sequence[TStr]]
            The tuple contains:
                - the new EnvID
                - the set of sources
                - the set of user dependencies
                - the exact set of dependencies
                - the set of extras
        """
        if architecture != base_env.env_id.arch:
            raise CondaException(
                "Mismatched architecture when extending an environment"
            )
        # We form the new list of dependencies based on the ones we have in cur_env
        # and the new ones.
        # We need to split things up and reform

        incoming_conda_deps = split_into_dict(
            [d for d in deps if d.category == "conda"]
        )
        incoming_npconda_deps = split_into_dict(
            [d for d in deps if d.category == "npconda"]
        )
        incoming_pip_deps = split_into_dict([d for d in deps if d.category == "pip"])

        user_conda_deps = merge_dep_dicts(
            split_into_dict([d for d in base_env.deps if d.category == "conda"]),
            incoming_conda_deps,
        )
        user_pip_deps = merge_dep_dicts(
            split_into_dict([d for d in base_env.deps if d.category == "pip"]),
            incoming_pip_deps,
        )
        user_npconda_deps = merge_dep_dicts(
            split_into_dict([d for d in base_env.deps if d.category == "npconda"]),
            incoming_npconda_deps,
        )

        # Extract the implicit channels that are already listed somewhere
        sources = list(set(chain(base_env.sources, sources)))
        extras = list(set(chain(base_env.extras, extras)))

        included_channels = set()  # type: Set[str]
        for s in chain(sources, conda.default_conda_channels):
            channel = channel_from_url(s.value if isinstance(s, TStr) else s)
            if channel:
                included_channels.add(channel)

        included_channels_list = list(included_channels)
        conda_deps = {
            p.package_name_with_channel(included_channels_list): p.package_version
            for p in base_env.packages
            if p.TYPE == "conda"
        }
        conda_deps = merge_dep_dicts(conda_deps, incoming_conda_deps)
        conda_deps = merge_dep_dicts(conda_deps, incoming_npconda_deps)

        pip_deps = {
            p.package_name: p.package_version
            for p in base_env.packages
            if p.TYPE == "pip"
        }
        pip_deps = merge_dep_dicts(pip_deps, incoming_pip_deps)

        deps = list(
            chain(
                [
                    TStr("conda", "%s==%s" % (k, v) if v else k)
                    for k, v in conda_deps.items()
                ],
                [
                    TStr("pip", "%s==%s" % (k, v) if v else k)
                    for k, v in pip_deps.items()
                ],
            )
        )

        user_deps = list(
            chain(
                [
                    TStr("conda", "%s==%s" % (k, v) if v else k)
                    for k, v in user_conda_deps.items()
                ],
                [
                    TStr("npconda", "%s==%s" % (k, v) if v else k)
                    for k, v in user_npconda_deps.items()
                ],
                [
                    TStr("pip", "%s==%s" % (k, v) if v else k)
                    for k, v in user_pip_deps.items()
                ],
            )
        )
        new_env_id = EnvID(
            ResolvedEnvironment.get_req_id(user_deps, sources, extras),
            "_default",
            architecture,
        )

        return new_env_id, sources, user_deps, deps, extras



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

import email.parser
import email.policy
import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile
import uuid

from enum import Enum
from itertools import chain
from typing import (
    Dict,
    FrozenSet,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)
from urllib.parse import urlparse, unquote

from requests import PreparedRequest
from requests.auth import AuthBase, HTTPBasicAuth

from metaflow_extensions.netflix_ext.vendor.packaging.tags import (
    compatible_tags,
    _cpython_abis,
    cpython_tags,
    mac_platforms,
    Tag,
)

from metaflow_extensions.netflix_ext.vendor.packaging.utils import (
    BuildTag,
    parse_wheel_filename,
)
from metaflow_extensions.netflix_ext.vendor.packaging.version import Version

from metaflow.debug import debug
from metaflow.exception import MetaflowException
import metaflow.metaflow_config as mf_config
from metaflow.metaflow_config import (
    CONDA_MAGIC_FILE_V2,  # type: ignore
    CONDA_PREFERRED_FORMAT,  # type: ignore
    CONDA_SRCS_AUTH_INFO,
)

from metaflow.metaflow_environment import InvalidEnvironmentException

# NOTA: Most of the code does not assume that there are only two formats BUT the
# transmute code does (since you can only specify the infile -- the outformat and file
# are inferred)
_ALL_CONDA_FORMATS = (".tar.bz2", ".conda")
# NOTE: Order is important as it is a preference order
_ALL_PIP_FORMATS = (".whl", ".tar.gz", ".zip")
_VALID_IMAGE_NAME = "[^-a-z0-9_/]"
_VALID_TAG_NAME = "[^-a-z0-9_]"


class AliasType(Enum):
    PATHSPEC = "pathspec"
    FULL_ID = "full-id"
    REQ_FULL_ID = "both-id"
    GENERIC = "generic"


# List of formats that guarantees the preferred format is first. This is important as
# functions that rely on selecting the "preferred" source of a package rely on the
# preferred format being first.
if CONDA_PREFERRED_FORMAT:
    CONDA_FORMATS = (
        CONDA_PREFERRED_FORMAT,
        *[x for x in _ALL_CONDA_FORMATS if x != CONDA_PREFERRED_FORMAT],
    )  # type: Tuple[str, ...]
else:
    CONDA_FORMATS = _ALL_CONDA_FORMATS  # type: Tuple[str, ...]
FAKEURL_PATHCOMPONENT = "_fake"


class CondaException(MetaflowException):
    headline = "Conda ran into an error while setting up environment."

    def __init__(self, error: Union[Sequence[Exception], str]):
        if isinstance(error, list):
            error = "\n".join([str(x) for x in error])
        super(CondaException, self).__init__(error)  # type: ignore


class CondaStepException(CondaException):
    def __init__(self, exception: CondaException, steps: Sequence[str]):
        msg = "Step(s): {steps}, Error: {error}".format(
            steps=steps, error=exception.message
        )
        super(CondaStepException, self).__init__(msg)


def convert_filepath(path: str, file_format: Optional[str] = None) -> Tuple[str, str]:
    if file_format and not path.endswith(file_format):
        path, cur_ext = correct_splitext(path)
        if cur_ext not in _ALL_CONDA_FORMATS:
            raise ValueError(
                "URL '%s' does not end with a supported file format %s"
                % (path, str(_ALL_CONDA_FORMATS))
            )
        path = "%s%s" % (path, file_format)
    return os.path.split(path)


def get_conda_manifest_path(ds_root: str) -> str:
    return os.path.join(ds_root, CONDA_MAGIC_FILE_V2)  # type: ignore


def get_conda_root(datastore_type: str) -> str:
    conda_root = getattr(mf_config, "CONDA_%sROOT" % datastore_type.upper())
    if conda_root is None:
        # We error on METAFLOW_DATASTORE_SYSROOT_<ds> because that is the default used
        raise MetaflowException(
            msg="METAFLOW_DATASTORE_SYSROOT_%s must be set!" % datastore_type.upper()
        )
    debug.conda_exec("Conda root is at %s" % conda_root)
    return conda_root  # type: ignore


def arch_id() -> str:
    bit = "32"
    if platform.machine().endswith("64"):
        bit = "64"
    if platform.system() == "Linux":
        return "linux-%s" % bit
    elif platform.system() == "Darwin":
        # Support M1 Mac
        if platform.machine() == "arm64":
            return "osx-arm64"
        else:
            return "osx-%s" % bit
    else:
        raise InvalidEnvironmentException(
            "The *@conda* decorator is not supported "
            "outside of Linux and Darwin platforms"
        )


def pip_tags_from_arch(python_version: str, arch: str) -> List[Tag]:
    # Converts a Conda architecture to a tuple containing (implementation, platforms, abis)
    # This function will assume a CPython implementation

    # This is inspired by what pip does:
    # https://github.com/pypa/pip/blob/0442875a68f19b0118b0b88c747bdaf6b24853ba/src/pip/_internal/utils/compatibility_tags.py
    py_version = tuple(map(int, python_version.split(".")[:2]))
    if arch == "linux-64":
        platforms = [
            "manylinux%s_x86_64" % s
            for s in (
                "1",
                "2010",
                "2014",
                "_2_17",
                "_2_18",
                "_2_19",
                "_2_20",
                "_2_21",
                "_2_23",
                "_2_24",
                "_2_25",
                "_2_26",
                "_2_27",
            )
        ]
        platforms.append("linux_x86_64")
    elif arch == "osx-64":
        platforms = mac_platforms((11, 0), "x86_64")
    elif arch == "osx-arm64":
        platforms = mac_platforms((11, 0), "arm64")
    else:
        raise InvalidEnvironmentException("Unsupported platform: %s" % arch)

    interpreter = "cp%s" % ("".join(map(str, py_version)))

    abis = _cpython_abis(py_version)

    supported = []  # type: List[Tag]
    supported.extend(cpython_tags(py_version, abis, platforms))
    supported.extend(compatible_tags(py_version, interpreter, platforms))
    return supported


ParseExplicitResult = NamedTuple(
    "ParseExplicitResult",
    [("filename", str), ("url", str), ("url_format", str), ("hash", Optional[str])],
)


def parse_explicit_url_conda(url: str) -> ParseExplicitResult:
    # Takes a URL in the form url#hash and returns:
    #  - the filename
    #  - the URL (without the hash)
    #  - the format for the URL
    #  - the hash
    filename = None
    url_format = None

    url_clean, url_hash = url.rsplit("#", 1)
    filename, url_format = correct_splitext(os.path.split(urlparse(url_clean).path)[1])
    if url_format not in _ALL_CONDA_FORMATS:
        raise CondaException(
            "URL '%s' is not a supported format (%s)" % (url, CONDA_FORMATS)
        )
    filename = unquote(filename)
    return ParseExplicitResult(
        filename=filename, url=url_clean, url_format=url_format, hash=url_hash
    )


def parse_explicit_path_pip(path: str) -> ParseExplicitResult:
    # Takes a filename in the form file://<path> and returns:
    #  - the filename
    #  - the URL (always file://local-<uuid>/<filename> so there is no way another
    #    build/user conflicts. We consider them to be all distinct.
    #  - the format of the URL
    #  - the hash will be set to None
    if not path.startswith("file://"):
        raise CondaException("Local path '%s' does not start with file://" % path)
    path = path[7:]
    orig_filename, url_format = correct_splitext(os.path.basename(path))
    if url_format not in _ALL_PIP_FORMATS:
        raise CondaException(
            "Path '%s' is not a supported format (%s)" % (path, str(_ALL_PIP_FORMATS))
        )
    return ParseExplicitResult(
        filename=unquote(orig_filename),
        url="file://local-%s/%s" % (str(uuid.uuid4()), orig_filename),
        url_format=url_format,
        hash=None,
    )


def parse_explicit_url_pip(url: str) -> ParseExplicitResult:
    # Takes a URL in the form url#hash and returns:
    #  - the filename
    #  - the URL (without the hash)
    #  - the format for the URL
    #  - the hash
    filename = None
    url_format = None

    url_clean, url_hash = url.rsplit("#", 1)
    if url_hash:
        if not url_hash.startswith("sha256="):
            raise CondaException("URL '%s' has a SHA type which is not supported" % url)
        url_hash = url_hash[7:]
    else:
        url_hash = None

    filename, url_format = correct_splitext(os.path.split(urlparse(url_clean).path)[1])
    if url_format not in _ALL_PIP_FORMATS:
        raise CondaException(
            "URL '%s' is not a supported format (%s)" % (url, str(_ALL_PIP_FORMATS))
        )
    filename = unquote(filename)
    return ParseExplicitResult(
        filename=filename, url=url_clean, url_format=url_format, hash=url_hash
    )


def plural_marker(count: int) -> str:
    return "s" if count != 1 else ""


def is_hexadecimal(s: str) -> bool:
    return not re.search("[^0-9a-f]", s)


def resolve_env_alias(env_alias: str) -> Tuple[AliasType, str]:
    if env_alias.startswith("step:"):
        env_alias = env_alias[5:]
        if len(env_alias.split("/")) == 3:
            return AliasType.PATHSPEC, env_alias
    elif len(env_alias) == 81 and env_alias[40] == ":":
        # req-id:full-id possibly
        req_id, full_id = env_alias.split(":", 1)
        if is_hexadecimal(req_id) and is_hexadecimal(full_id):
            return AliasType.REQ_FULL_ID, env_alias
    elif len(env_alias) == 40 and is_hexadecimal(env_alias):
        # For now we do not support this -- remove if you want to support
        # The issue with supporting this is that a full-id can refer to multiple req-id
        # so it is impossible to unambiguously identify the source environment (ie:
        # the user requested dependencies, etc)
        raise MetaflowException(
            "Invalid format for environment alias: '%s'" % env_alias
        )
        # return AliasType.FULL_ID, env_alias
    else:
        splits = env_alias.rsplit(":", 1)
        if len(splits) == 2:
            image_name = splits[0]
            image_tag = splits[1]
        else:
            image_name = env_alias
            image_tag = "latest"
        if re.search(_VALID_IMAGE_NAME, image_name):
            raise MetaflowException(
                "An environment name must contain only "
                "lowercase alphanumeric characters, dashes, underscores and forward slashes."
            )
        if image_name[0] == "/" or image_name[-1] == "/":
            raise MetaflowException(
                "An environment name must not start or end with '/'"
            )
        if re.search(_VALID_TAG_NAME, image_tag):
            raise MetaflowException(
                "An environment tag name must contain only "
                "lowercase alphanumeric characters, dashes and underscores."
            )
        return AliasType.GENERIC, "/".join([image_name, image_tag])
    raise MetaflowException("Invalid format for environment alias: '%s'" % env_alias)


def is_alias_mutable(alias_type: AliasType, resolved_alias: str) -> bool:
    if alias_type != AliasType.GENERIC:
        return False
    splits = resolved_alias.rsplit("/", 1)
    return len(splits) == 2 and splits[1] in ("latest", "candidate", "stable")


def split_into_dict(deps: List["TStr"]) -> Dict[str, str]:
    result = {}  # type: Dict[str, str]
    for dep in deps:
        s = dep.value.split("==", 1)
        if len(s) == 1:
            result[s[0]] = ""
        else:
            result[s[0]] = s[1]
    return result


def merge_dep_dicts(
    d1: Dict[str, str], d2: Dict[str, str], only_last_deps: bool = False
) -> Dict[str, str]:
    # Merge dictionaries of version constraints taking *all* constraints
    # into account instead of replacing them.
    # Order will be preserved based on the order of the first dict
    # only_last_deps will only return things that are already in d2. This is to merge with
    # anything in d1 but only return things that are in d2. This is used to compute
    # the new set of dependencies introduced by the user.
    result = {}  # type: Dict[str, str]
    for key in set(d2.keys() if only_last_deps else chain(d1.keys(), d2.keys())):
        v1 = [v for v in d1.get(key, "").split(",") if v]
        v2 = [v for v in d2.get(key, "").split(",") if v]
        sv1 = set([v.lstrip("=") for v in v1])
        # We dedup so that if we have the exact same constraints, we don't change
        # things. We try to keep the order of the first dict
        v2 = [v for v in v2 if v.lstrip("=") not in sv1]

        # We need to make sure that things that don't start with ~, <, >, ! or = get
        # == if they are not first.
        resulting_versions = list(chain(v1, v2))
        if resulting_versions:
            resulting_versions = [resulting_versions[0]] + [
                v if v[0] in ("~", "<", ">", "!", "=") else "==%s" % v
                for v in resulting_versions[1:]
            ]
        result[key] = ",".join(resulting_versions)
    return result


def reform_pip_filename(
    name: str, version: Version, build: BuildTag, tags: FrozenSet[Tag]
) -> str:
    if build:
        build_str = "%d%s" % (build[0], build[1])
    else:
        build_str = None
    # We need to iterate over all the tags to form the string
    # interp1.interp2-abi1.abi2-platform1.platform2
    interpreters, abis, platforms = zip(
        *map(lambda x: (x.interpreter, x.abi, x.platform), tags)
    )
    tag_str = "-".join(
        [
            ".".join(sorted(interpreters)),
            ".".join(sorted(abis)),
            ".".join(sorted(platforms)),
        ]
    )

    if build_str:
        return "-".join([str(name), str(version), build_str, tag_str])
    return "-".join([str(name), str(version), tag_str])


def correct_splitext(filename_with_ext: str) -> Tuple[str, str]:
    # This handles .tar.gz and .tar.bz2 which is what we care about.
    # Tried with pathlib too but that has other issues since . is a valid character
    # in conda packages
    all_splits = filename_with_ext.rsplit(".", 2)
    if len(all_splits) > 2:
        if ".".join(all_splits[1:]) in ("tar.gz", "tar.bz2"):
            return all_splits[0], ".".join([""] + all_splits[1:])
        return ".".join(all_splits[:-1]), "." + all_splits[2]
    if len(all_splits) == 2:
        return all_splits[0], "." + all_splits[1]
    if len(all_splits) == 1:
        return filename_with_ext, ""
    # Should never happen
    return "", ""


def conda_deps_to_pip_deps(d: Dict[str, str]) -> Dict[str, str]:
    # This is a hack for now -- real fix would be to specify get_pinned_pip_libs
    # but at least for now this allows channels to be specified in the conda_libs
    # but still work with pip
    r = {}  # type: Dict[str, str]
    for k, v in d.items():
        s = k.split("::")
        if len(s) > 1:
            r[s[1]] = v
        else:
            r[k] = v
    return r


_UNDERSCORE_REGEX = re.compile(r"[-_.]+")


def normalize_to_underscore(name: str) -> str:
    return _UNDERSCORE_REGEX.sub("_", name).lower()


def channel_from_url(url: str) -> Optional[str]:
    up = urlparse(url)
    if up.hostname == "conda.anaconda.org":
        return up.path.split("/", 2)[1]
    return None


def auth_from_urls(urls: List[str]) -> Optional[AuthBase]:
    auths_per_hostname = {
        cast(str, h): HTTPBasicAuth(username, pwd)
        for h, (username, pwd) in CONDA_SRCS_AUTH_INFO.items()
    }

    for url in urls:
        up = urlparse(url.strip("'\" \t\n"))
        if up.hostname and up.username:
            auths_per_hostname[up.hostname] = HTTPBasicAuth(
                up.username, up.password or ""
            )
    if auths_per_hostname:
        return PerHostAuth(auths_per_hostname)
    return None


class PerHostAuth(AuthBase):
    def __init__(self, auths_per_hostname: Mapping[str, AuthBase]):
        self._my_auths = auths_per_hostname

    def __call__(self, r: PreparedRequest):
        up = urlparse(r.url)
        if up.hostname:
            h = (
                up.hostname
                if isinstance(up.hostname, str)
                else up.hostname.decode(encoding="utf-8")
            )
            auth_provider = self._my_auths.get(h or "<none>")
            if auth_provider:
                return auth_provider(r)
        return r


# Function heavily inspired from https://github.com/hauntsaninja/change_wheel_version
# MIT license of that source file:
# MIT License

# Copyright (c) 2023 hauntsaninja

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
def change_pip_package_version(wheel_path: str, new_version_str: str) -> str:
    base_dir, wheel_name = os.path.split(wheel_path)
    with tempfile.TemporaryDirectory() as build_dir:
        old_parts = parse_wheel_filename(wheel_name)

        distribution = wheel_name.split("-", 1)[0]
        old_version = old_parts[1]
        new_version = Version(new_version_str)
        new_version_str = str(new_version)

        old_slug = "%s-%s" % (distribution, old_version)
        new_slug = "%s-%s" % (distribution, new_version)

        subprocess.check_output(
            [sys.executable, "-m", "wheel", "unpack", "-d", build_dir, wheel_path]
        )
        if not os.path.isdir(os.path.join(build_dir, old_slug)) or not os.path.isdir(
            os.path.join(build_dir, old_slug, "%s.dist-info" % old_slug)
        ):
            raise ValueError("Invalid wheel expansion for '%s'" % wheel_path)

        # Copy everything
        shutil.move(
            os.path.join(build_dir, old_slug), os.path.join(build_dir, new_slug)
        )

        # Rename dist-info directory
        shutil.move(
            os.path.join(build_dir, new_slug, "%s.dist-info" % old_slug),
            os.path.join(build_dir, new_slug, "%s.dist-info" % new_slug),
        )

        # Rename data if it exists
        if os.path.exists(os.path.join(build_dir, new_slug, "%s.data" % old_slug)):
            shutil.move(
                os.path.join(build_dir, new_slug, "%s.data" % old_slug),
                os.path.join(build_dir, new_slug, "%s.data" % new_slug),
            )

        # Rewrite METADATA file
        metadata = os.path.join(
            build_dir, new_slug, "%s.dist-info" % new_slug, "METADATA"
        )

        # This is actually a non-conformant email policy as per
        # https://packaging.python.org/en/latest/specifications/core-metadata/
        # However, it works around this bug in setuptools in cases where the version is really long
        # https://github.com/pypa/setuptools/issues/3808
        max_line_length = 200
        policy = email.policy.compat32.clone(max_line_length=200)
        if len(new_version_str) >= max_line_length:
            raise ValueError(f"Version '%s' is too long" % new_version_str)

        with open(metadata, "rb") as f:
            parser = email.parser.BytesParser(policy=policy)
            msg = parser.parse(f)

        msg.replace_header("Version", new_version_str)
        with open(metadata, "wb") as f:
            f.write(msg.as_bytes())

        # Rewrites the RECORD file and generate the file.
        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "wheel",
                "pack",
                "-d",
                build_dir,
                os.path.join(build_dir, new_slug),
            ]
        )

        # We check if we get the expected name
        expected_name = (
            reform_pip_filename(distribution, new_version, old_parts[2], old_parts[3])
            + ".whl"
        )
        if os.path.isfile(os.path.join(build_dir, expected_name)):
            shutil.move(
                os.path.join(build_dir, expected_name),
                os.path.join(base_dir, expected_name),
            )
            os.unlink(wheel_path)
            return os.path.join(base_dir, expected_name)
        raise RuntimeError(
            "Could not rename wheel '%s'; expected '%s' got: %s"
            % (wheel_path, expected_name, ", ".join(os.listdir(build_dir)))
        )



from metaflow.decorators import FlowDecorator
from metaflow.metaflow_environment import InvalidEnvironmentException


class CondaFlowDecorator(FlowDecorator):
    """
    Specifies the Conda environment for all steps of the flow.

    Use `@conda_base` to set common libraries required by all
    steps and use `@conda` to specify step-specific additions.

    Parameters
    ----------
    name : Optional[str]
        If specified, can refer to a named environment. The environment referred to
        here will be the one used as a base environment for all steps.
        If specified, nothing else can be specified in this decorator.
    pathspec : Optional[str]
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used as a base environment for all steps.
        If specified, nothing else can be specified in this decorator.
    libraries : Optional[Dict[str, str]]
        Libraries to use for this step. The key is the name of the package
        and the value is the version to use (default: `{}`). Note that versions can
        be specified either as a specific version or as a comma separated string
        of constraints like "<2.0,>=1.5".
    channels : Optional[List[str]]
        Additional channels to search
    pip_packages : Optional[Dict[str, str]]
        Same as libraries but for pip packages.
    pip_sources : Optional[List[str]]
        Same as channels but for pip sources.
    python : Optional[str]
        Version of Python to use, e.g. '3.7.4'. If not specified, the current Python
        version will be used.
    fetch_at_exec : bool, default False
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this step to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    disabled : bool, default False
        If set to True, disables Conda
    """

    name = "conda_base"
    defaults = {
        "name": None,
        "pathspec": None,
        "libraries": {},
        "channels": [],
        "pip_packages": {},
        "pip_sources": [],
        "python": None,
        "fetch_at_exec": None,
        "disabled": None,
    }

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        if "pip_base" in flow._flow_decorators:
            raise InvalidEnvironmentException(
                "conda_base decorator is not compatible with pip_base. "
                "Please specify only one of them."
            )
        if environment.TYPE != "conda":
            raise InvalidEnvironmentException(
                "The *@conda_base* decorator requires " "--environment=conda"
            )

        if (self.attributes["name"] or self.attributes["pathspec"]) and len(
            [
                k
                for k, v in self.attributes.items()
                if v and k not in ("name", "pathspec")
            ]
        ):
            raise InvalidEnvironmentException(
                "You cannot specify `name` or `pathspec` along with other attributes in @%s"
                % self.name
            )


class PipFlowDecorator(FlowDecorator):
    """
    Specifies the Pip environment for all steps of the flow.

    Use `@pip_base` to set common libraries required by all
    steps and use `@pip` to specify step-specific additions.

    Parameters
    ----------
    name : Optional[str]
        If specified, can refer to a named environment. The environment referred to
        here will be the one used as a base environment for all steps.
        If specified, nothing else can be specified in this decorator.
    pathspec : Optional[str]
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used as a base environment for all steps.
        If specified, nothing else can be specified in this decorator.
    packages : Optional[Dict[str, str]]
        Packages to use for this step. The key is the name of the package
        and the value is the version to use (default: `{}`). Note that versions can
        be specified either as a specific version or as a comma separated string
        of constraints like "<2.0,>=1.5".
    sources : Optional[List[str]]
        Additional channels to search for
    python : Optional[str]
        Version of Python to use, e.g. '3.7.4'. If not specified, the current version
        will be used.
    fetch_at_exec : bool, default False
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this step to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    disabled : bool, default False
        If set to True, disables Pip
    """

    name = "pip_base"

    defaults = {
        "name": None,
        "pathspec": None,
        "packages": {},
        "sources": [],
        "python": None,
        "fetch_at_exec": None,
        "disabled": None,
    }

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        if "conda_base" in flow._flow_decorators:
            raise InvalidEnvironmentException(
                "pip_base decorator is not compatible with conda_base. "
                "Please specify only one of them."
            )
        if environment.TYPE != "conda":
            raise InvalidEnvironmentException(
                "The *@pip_base* decorator requires --environment=conda"
            )

        if (self.attributes["name"] or self.attributes["pathspec"]) and len(
            [
                k
                for k, v in self.attributes.items()
                if v and k not in ("name", "pathspec")
            ]
        ):
            raise InvalidEnvironmentException(
                "You cannot specify `name` or `pathspec` along with other attributes in @%s"
                % self.name
            )



__mf_extensions__ = "netflix-ext"

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("metaflow-netflixext").version
except:
    # this happens on remote environments since the job package
    # does not have a version
    __version__ = None
